"""
This module does a one-way-sync between azure and databricks users and groups.
Author: nicolas.jankelevitch@dataroots.io
"""


from databricks_azure_ad_sync_provider.sync_classes import SyncGroup, SyncServicePrincipal, SyncStatus, SyncUser


class OneWayMatch:
    def __init__(self, azure_details, databricks_identity):
        self.databricks_identity = databricks_identity
        self.azure_groups = azure_details["members_group"]
        self.azure_users = azure_details["members_user"]
        self.azure_sps = azure_details["members_sp"]
        self.azure_member_structure = azure_details["members_structure"]

    def get_sync_group(
        self, azure_group, databricks_group, status: SyncStatus
    ) -> SyncGroup:
        """
        This method takes info from Azure and Databricks groups and combine them
        Return: A SyncGroup object
        """
        sg = SyncGroup(azure_group, databricks_group, self.databricks_identity)
        sg.status = status

        if azure_group["id"]:
            sg.azure_members = self.azure_member_structure[azure_group["id"]]
        return sg

    def match_groups(self):
        """
        This method checks if there is a match between Azure AD and Databricks groups

        Return: A list of SyncGroup objects to be synced
        """
        sync_groups = []
        databricks_dummy_group = {
            "id": "",
            "display_name": "",
            "external_id": "",
            "members": [],
        }
        azure_dummy_group = {"id": "", "display_name": ""}

        all_databricks_groups = self.databricks_identity.all_databricks_objects[
            "groups"
        ]
        databricks_groups_id = [x for x in all_databricks_groups]
        for azure_group in self.azure_groups:
            # check for match on id
            match = [
                {
                    "id": x,
                    "display_name": all_databricks_groups[x]["display_name"],
                    "external_id": all_databricks_groups[x]["external_id"],
                    "members": all_databricks_groups[x]["members"],
                }
                for x in all_databricks_groups
                if all_databricks_groups[x]["external_id"] == azure_group["id"]
                or (
                    x in self.databricks_identity.relations["groups"]
                    and self.databricks_identity.relations["groups"][x]
                    == azure_group["id"]
                )
                or all_databricks_groups[x]["display_name"]
                == azure_group["display_name"]
            ]

            if len(match) > 1:
                raise Exception(
                    f"Multiple matches found for group with ID {azure_group['id']}."
                )
            if match:
                # As long as we find a match, we store the relation in the json file
                self.databricks_identity.relations["groups"][
                    match[0]["id"]
                ] = azure_group["id"]
                m = self.get_sync_group(azure_group, match[0], SyncStatus.TO_UPDATE)
                databricks_groups_id.remove(match[0]["id"])
                if not m.is_in_sync():
                    sync_groups.append(m)
            else:
                m = self.get_sync_group(
                    azure_group, databricks_dummy_group, SyncStatus.TO_ADD
                )
                sync_groups.append(m)

        for db in databricks_groups_id:
            d = self.get_sync_group(
                azure_dummy_group,
                {
                    "id": db,
                    "display_name": all_databricks_groups[db]["display_name"],
                    "external_id": all_databricks_groups[db]["external_id"],
                    "members": all_databricks_groups[db]["members"],
                },
                SyncStatus.TO_DELETE,
            )
            sync_groups.append(d)

        return sync_groups

    def get_sync_user(self, azure_user, databricks_user, status) -> SyncUser:
        """
        This method takes info from Azure and Databricks users and combine them
        Return: A SyncUser object
        """
        su = SyncUser(azure_user, databricks_user, self.databricks_identity)
        su.status = status
        return su

    def match_users(self) -> list[SyncUser]:
        """
        This method checks if there is a match between Azure AD and Databricks users

        Return: A list of SyncUser objects to be synced
        """
        sync_users = []
        databricks_dummy_user = {
            "id": "",
            "display_name": "",
            "external_id": "",
            "user_name": "",
        }
        azure_dummy_user = {"id": "", "display_name": "", "user_principal_name": ""}
        all_databricks_users = self.databricks_identity.all_databricks_objects["users"]
        databricks_users_id = [x for x in all_databricks_users]
        for azure_user in self.azure_users:
            # check for match on id
            match = [
                {
                    "id": x,
                    "display_name": all_databricks_users[x]["display_name"],
                    "external_id": all_databricks_users[x]["external_id"],
                    "user_name": all_databricks_users[x]["user_name"],
                }
                for x in all_databricks_users
                if all_databricks_users[x]["external_id"] == azure_user["id"]
                # check here for match between AZ ID and databricks ID
                or (
                    x in self.databricks_identity.relations["users"]
                    and self.databricks_identity.relations["users"][x]
                    == azure_user["id"]
                )
                or all_databricks_users[x]["user_name"]
                == azure_user["user_principal_name"]
                or all_databricks_users[x]["display_name"] == azure_user["display_name"]
            ]

            if len(match) > 1:
                raise Exception(
                    f"Multiple matches found for user with ID {azure_user['id']}."
                )
            if match:
                self.databricks_identity.relations["users"][
                    match[0]["id"]
                ] = azure_user["id"]
                m = self.get_sync_user(azure_user, match[0], SyncStatus.TO_UPDATE)
                databricks_users_id.remove(match[0]["id"])
                if not m.is_in_sync():
                    sync_users.append(m)
            else:
                m = self.get_sync_user(
                    azure_user, databricks_dummy_user, SyncStatus.TO_ADD
                )
                sync_users.append(m)
        for du in databricks_users_id:
            sync_users.append(
                self.get_sync_user(
                    azure_dummy_user,
                    {
                        "id": du,
                        "display_name": all_databricks_users[du]["display_name"],
                        "external_id": all_databricks_users[du]["external_id"],
                        "user_name": all_databricks_users[du]["user_name"],
                    },
                    SyncStatus.TO_DELETE,
                )
            )
        return sync_users

    def get_sync_service_principal(
        self, azure_sp, databricks_sp, status
    ) -> SyncServicePrincipal:
        """
        This method takes info from Azure and Databricks sps and combine them
        Return: A SyncServicePrincipal object
        """
        ssp = SyncServicePrincipal(azure_sp, databricks_sp, self.databricks_identity)
        ssp.status = status
        return ssp

    def match_service_principals(self) -> list[SyncServicePrincipal]:
        """
        This method checks if there is a match between Azure AD and Databricks sps

        Return: A list of SyncServicePrincipal objects to be synced
        """
        sync_sps = []

        databricks_dummy_sp = {
            "id": "",
            "display_name": "",
            "external_id": "",
            "application_id": "",
        }
        azure_dummy_sp = {"id": "", "display_name": "", "app_id": ""}
        all_databricks_sps = self.databricks_identity.all_databricks_objects["sps"]
        databricks_sps_id = [x for x in all_databricks_sps]

        for azure_sp in self.azure_sps:
            # check for match on id
            match = [
                {
                    "id": x,
                    "display_name": all_databricks_sps[x]["display_name"],
                    "external_id": all_databricks_sps[x]["external_id"],
                    "application_id": all_databricks_sps[x]["application_id"],
                }
                for x in all_databricks_sps
                if all_databricks_sps[x]["external_id"] == azure_sp["id"]
                or (
                    x in self.databricks_identity.relations["servicePrincipals"]
                    and self.databricks_identity.relations["servicePrincipals"][x]
                    == azure_sp["id"]
                )
                or all_databricks_sps[x]["display_name"] == azure_sp["display_name"]
                or all_databricks_sps[x]["application_id"] == azure_sp["app_id"]
            ]

            if len(match) > 1:
                raise Exception(
                    f"""Multiple matches found for
                    service principal with ID {azure_sp['id']}."""
                )
            if match:
                self.databricks_identity.relations["servicePrincipals"][
                    match[0]["id"]
                ] = azure_sp["id"]
                m = self.get_sync_service_principal(
                    azure_sp, match[0], SyncStatus.TO_UPDATE
                )
                databricks_sps_id.remove(match[0]["id"])
                if not m.is_in_sync():
                    sync_sps.append(m)
            else:
                m = self.get_sync_service_principal(
                    azure_sp, databricks_dummy_sp, SyncStatus.TO_ADD
                )
                sync_sps.append(m)

        for du in databricks_sps_id:
            sync_sps.append(
                self.get_sync_service_principal(
                    azure_dummy_sp,
                    {
                        "id": du,
                        "display_name": all_databricks_sps[du]["display_name"],
                        "external_id": all_databricks_sps[du]["external_id"],
                        "application_id": all_databricks_sps[du]["application_id"],
                    },
                    SyncStatus.TO_DELETE,
                )
            )

        return sync_sps



"""
This module is used for all Sync classes
"""
from enum import Enum


class SyncStatus(Enum):
    INITIAL = 0
    TO_ADD = 1
    TO_DELETE = 2
    TO_UPDATE = 3


class SyncUser:
    def __init__(self, azure_user, databricks_user, databricks_identity):
        self.databricks_display_name = databricks_user["display_name"]
        self.databricks_id = databricks_user["id"]
        self.databricks_external_id = databricks_user["external_id"]
        self.databricks_email = databricks_user["user_name"]

        self.azure_id = azure_user["id"]
        self.azure_display_name = azure_user["display_name"]
        self.azure_email = azure_user["user_principal_name"]
        self.status = SyncStatus.INITIAL
        self.databricks_identity = databricks_identity

    def is_in_sync(self) -> bool:
        """

        Returns:
            bool: True if the Databricks and Azure user are already synced
        """
        display_name = self.databricks_display_name == self.azure_display_name
        email = self.databricks_email == self.azure_email
        id = (
            self.databricks_external_id == self.azure_id
            or self.databricks_identity.relations["users"][self.databricks_id]
            == self.azure_id
        )
        return display_name and email and id

    def sync_user_to_db(self, force_delete=False):
        """Sync a users to Databricks
        Args:
            force_delete : If True, allows to delete users in Databricks
        """
        id = self.databricks_id
        user_name = self.azure_email
        display_name = self.azure_display_name
        external_id = self.azure_id

        if self.status == SyncStatus.TO_UPDATE:
            print(f"Updating user {display_name}")
            self.databricks_identity.update_databricks_user(
                id, display_name, user_name, external_id
            )

        elif self.status == SyncStatus.TO_ADD:
            print(f"Adding user {display_name}")
            self.databricks_identity.add_databricks_user(
                display_name, user_name, external_id
            )

        elif self.status == SyncStatus.TO_DELETE:
            if force_delete:
                self.databricks_identity.delete_databricks_user(id)

        else:
            pass


class SyncGroup:
    def __init__(self, azure_group, databricks_group, databricks_identity):
        self.databricks_id = databricks_group["id"]
        self.databricks_external_id = databricks_group["external_id"]
        self.databricks_display_name = databricks_group["display_name"]
        self.databricks_members = databricks_group["members"]
        self.azure_id = azure_group["id"]
        self.azure_display_name = azure_group["display_name"]
        self.azure_members = []
        self.status = SyncStatus.INITIAL
        self.databricks_identity = databricks_identity

    def is_in_sync(self) -> bool:
        """

        Returns:
            bool: True if the Databricks and Azure groups are already synced
        """
        if len(self.databricks_members) != len(self.azure_members):
            return False

        i = (
            self.databricks_external_id == self.azure_id
            or self.databricks_identity.relations["groups"][self.databricks_id]
            == self.azure_id
        )
        display_name = self.databricks_display_name == self.azure_display_name

        azure_member_ids = [x["id"] for x in self.azure_members]
        databricks_member_ids = []
        for db_member in self.databricks_members:
            db_id = db_member.value
            if db_id in self.databricks_identity.relations["users"]:
                databricks_member_ids.append(
                    self.databricks_identity.relations["users"][db_id]
                )
            elif db_id in self.databricks_identity.relations["groups"]:
                databricks_member_ids.append(
                    self.databricks_identity.relations["groups"][db_id]
                )
            elif db_id in self.databricks_identity.relations["servicePrincipals"]:
                databricks_member_ids.append(
                    self.databricks_identity.relations["servicePrincipals"][db_id]
                )

        members = sorted(azure_member_ids) == sorted(databricks_member_ids)

        return i and display_name and members

    def sync_group_add_empty(self):
        """
        Add a new group without any group members
        """
        print(f"Adding group {self.azure_display_name}")
        # create a new group without any members
        new_group = self.databricks_identity.add_databricks_group(
            self.azure_display_name, self.azure_id
        )
        # We will update the group's members afterwards
        self.status = SyncStatus.TO_UPDATE
        self.databricks_id = new_group.id
        self.databricks_display_name = new_group.display_name
        self.databricks_external_id = new_group.external_id

    def sync_group_update(self, force_delete=False):
        """
        Sync a group to Databricks
        Args:
            force_delete : If True, allows to delete groups in Databricks
        """
        members = [x["id"] for x in self.azure_members]
        members = []
        all_relations = (
            self.databricks_identity.relations["users"]
            | self.databricks_identity.relations["groups"]
            | self.databricks_identity.relations["servicePrincipals"]
        )
        for x in self.azure_members:
            external_id = x["id"]
            for i in all_relations:
                if all_relations[i] == external_id:
                    members.append(i)
        if self.status == SyncStatus.TO_UPDATE:
            print(f"Updating group {self.azure_display_name}")
            self.databricks_identity.update_databricks_group(
                self.databricks_id, self.azure_display_name, members, self.azure_id
            )

        elif self.status == SyncStatus.TO_DELETE:
            if force_delete:
                self.databricks_identity.delete_databricks_user(self.databricks_id)

        else:
            pass


class SyncServicePrincipal:
    def __init__(
        self, azure_service_principal, databricks_service_principal, databricks_identity
    ):
        self.databricks_application_id = databricks_service_principal["application_id"]
        self.databricks_display_name = databricks_service_principal["display_name"]
        self.databricks_external_id = databricks_service_principal["external_id"]
        self.databricks_id = databricks_service_principal["id"]

        self.azure_id = azure_service_principal["id"]
        self.azure_display_name = azure_service_principal["display_name"]
        self.azure_application_id = azure_service_principal["app_id"]

        self.status = SyncStatus.INITIAL
        self.databricks_identity = databricks_identity

    def is_in_sync(self) -> bool:
        """

        Returns:
            bool: True if the Databricks and Azure user are already synced.
            Which means they have the same app_id, display_name and id
        """
        app_id = self.databricks_application_id == self.azure_application_id
        display_name = self.databricks_display_name == self.azure_display_name
        id = (
            self.databricks_external_id == self.azure_id
            or self.databricks_identity.relations["servicePrincipals"][
                self.databricks_id
            ]
            == self.azure_id
        )

        return app_id and display_name and id

    def sync_sp_to_db(self, force_delete=False):
        """Sync a Service Principal to Databricks
        Args:
            force_delete : If True, allows to delete service principals in Databricks
        """
        # do we need the external ID here as we already have an application ID ?
        id = self.databricks_id
        display_name = self.azure_display_name
        application_id = self.azure_application_id
        external_id = self.azure_id

        if self.status == SyncStatus.TO_UPDATE:
            print(f"Updating sp {display_name}")
            self.databricks_identity.update_databricks_sps(
                id, application_id, display_name, external_id
            )

        if self.status == SyncStatus.TO_ADD:
            print(f"Adding sp {display_name}")
            self.databricks_identity.add_databricks_sps(
                application_id, display_name, external_id
            )

        elif self.status == SyncStatus.TO_DELETE:
            if force_delete:
                self.databricks_identity.delete_databricks_user(id)

        else:
            pass



"""
This module is for Microsoft Entra ID (formely Azure Active Directory) objects
Author: xuyao@dataroots.io
"""
from msgraph import GraphServiceClient


class AzureIdentity:
    def __init__(self, azure_credential):
        """Authenticate AD service principle

        Args: client_id, client_secret, tenant_id
        """
        self.credential = azure_credential
        self.scopes = ["https://graph.microsoft.com/.default"]
        self.client = GraphServiceClient(
            credentials=self.credential, scopes=self.scopes
        )

    async def get_group_direct_members(
        self, group_object_id: str, exclude_objects: list[str]
    ) -> (list[dict], list[dict], list[dict]):
        """
        This method fetches all direct AD identities in a group

        Args: group_object_id
        return: a list of direct AD users
                a list of direct AD groups
                a list of direct AD service principals
        """

        url = "https://graph.microsoft.com/beta/groups/" + group_object_id + "/members"
        users_list = []
        groups_list = []
        sps_list = []

        while url:
            """
            This loop solves the paging problem
            """
            try:
                members = (
                    await self.client.groups.with_url(url)
                    .by_group_id(group_object_id)
                    .members.get()
                )
                for user in members.value:
                    if user.id not in exclude_objects:
                        if user.odata_type == "#microsoft.graph.user":
                            users_list.append(
                                {
                                    "id": user.id,
                                    "display_name": user.display_name,
                                    "user_principal_name": user.user_principal_name,
                                }
                            )
                        elif user.odata_type == "#microsoft.graph.group":
                            groups_list.append(
                                {
                                    "id": user.id,
                                    "display_name": user.display_name,
                                }
                            )
                        elif user.odata_type == "#microsoft.graph.servicePrincipal":
                            sps_list.append(
                                {
                                    "id": user.id,
                                    "display_name": user.display_name,
                                    "app_id": user.app_id,
                                }
                            )
                url = members.odata_next_link
            except ValueError:
                break

        return (users_list, groups_list, sps_list)

    async def get_group_members_recursively(
        self, group_object_id: str, exclude_objects: list[str]
    ) -> dict:
        """
        This method fetches all AD identities in a group recursively

        Args: group_object_id
              exclude_objects: a list of Azure object id to be excluded
        return: a dictionary containing the group info and its 3 types of members
        """
        root_group = await self.client.groups.by_group_id(group_object_id).get()
        groups_list = [group_object_id]
        group_members_users = []
        group_members_groups = [
            {
                "id": group_object_id,
                "display_name": root_group.display_name,
            }
        ]
        group_member_sps = []
        members_structure = {}

        while len(groups_list):
            group_id = groups_list.pop(0)
            (
                direct_users,
                direct_groups,
                direct_sps,
            ) = await self.get_group_direct_members(group_id, exclude_objects)

            members_structure[group_id] = direct_users + direct_groups + direct_sps

            group_members_users += direct_users
            group_members_groups += direct_groups
            group_member_sps += direct_sps

            for g in direct_groups:
                groups_list.append(g["id"])

        return {
            "members_user": group_members_users,
            "members_group": group_members_groups,
            "members_sp": group_member_sps,
            "members_structure": members_structure,
        }

    async def get_azure_objects_with_exclude(
        self, group_objects: list[str], exclude_objects: list[str]
    ) -> dict:
        """
        Args:
          group_objects: A list of Azure group object id
          exclude_objects: A list of Azure object ids which should be
                           excluded from the sync

        Return: A dict containing all users, groups, service principals to be synced
        """
        members_user = []
        members_group = []
        members_sp = []
        members_structure = {}

        for group_id in group_objects:
            members = await self.get_group_members_recursively(
                group_id, exclude_objects
            )
            members_user += members["members_user"]
            members_group += members["members_group"]
            members_sp += members["members_sp"]
            for g in members["members_structure"]:
                if g not in members_structure:
                    members_structure[g] = members["members_structure"][g]

        # drop duplicates
        members_user = {frozenset(item.items()): item for item in members_user}.values()
        members_group = {
            frozenset(item.items()): item for item in members_group
        }.values()
        members_sp = {frozenset(item.items()): item for item in members_sp}.values()

        return {
            "members_user": members_user,
            "members_group": members_group,
            "members_sp": members_sp,
            "members_structure": members_structure,
        }



import pytest

from src.databricks_azure_ad_sync_provider.yaml_parser import (
    is_azure_object_id,
    parse_yaml,
)


def test_raises_typeerror_when_no_groups_list():
    with pytest.raises(TypeError):
        parse_yaml("tests/test_files/no_groups_list.yaml")


@pytest.mark.parametrize(
    "object_id,result", [("123", False), ("cac07e36-df30-457a-bc31-4e92d71d07c5", True)]
)
def test_is_azure_object_id(object_id, result):
    assert is_azure_object_id(object_id) == result


@pytest.mark.parametrize(
    "filepath",
    ["invalid_object_id_in_groups.yaml", "invalid_object_id_in_exclude.yaml"],
)
def test_raises_valueerror_when_file_contains_invalid_object_id(filepath):
    with pytest.raises(ValueError):
        parse_yaml(f"tests/test_files/{filepath}")



"""
This module is the main logic to sync objects from Azure AD to Databricks
"""
import argparse
import asyncio
import logging
import sys
from pathlib import Path

from azure.identity import DefaultAzureCredential
from databricks_azure_ad_sync_provider.azure_identity import AzureIdentity
from databricks.sdk import AccountClient
from databricks_azure_ad_sync_provider.databricks_identity import DatabricksIdentity
from databricks_azure_ad_sync_provider.one_way_match import OneWayMatch
from databricks_azure_ad_sync_provider.sync_classes import SyncStatus
from databricks_azure_ad_sync_provider.yaml_parser import parse_yaml

json_path = (Path(__file__).parents[2] / "./syncstates.json").resolve()


logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.DEBUG,
    datefmt="%Y-%m-%d %H:%M:%S",
)

# parse arguments
parser = argparse.ArgumentParser()
parser.add_argument(
    "--file",
    "-f",
    type=str,
    required=True,
    help="The path to the yaml file",
)
parser.add_argument(
    "--delete",
    "-d",
    help="If specified, you enable databricks identities to be deleted",
    action="store_true",
)

args = parser.parse_args()
yaml_path = args.file
force_delete = args.delete

group_objects, exclude_objects = parse_yaml(yaml_path)

try:
    logging.info("Attempting to authenticate to Azure...")
    azure_credential = DefaultAzureCredential()
    azure_identity = AzureIdentity(azure_credential)
    logging.info("Authenticated to Azure.")

    logging.info("Attempting to authenticate to databricks...")
    db_account_client = AccountClient()
    logging.info("Authenticated to databricks.")

except ValueError as ve:
    logging.error(ve)
    sys.exit()

logging.info("Fetching databricks identities...")
databricks_identity = DatabricksIdentity(db_account_client, json_path)

logging.info("Calling Azure API...")
azure_group_details = asyncio.run(
    azure_identity.get_azure_objects_with_exclude(group_objects, exclude_objects)
)

ows = OneWayMatch(azure_group_details, databricks_identity)

logging.info("Compare and get users to be synced...")

sync_users = ows.match_users()

print("#" * 80)
print("USERS TO SYNC")
print("#" * 80)
for user in sync_users:
    print(f"{user.azure_display_name}, {user.status}")
    if user.status != SyncStatus.TO_DELETE:
        user.sync_user_to_db(force_delete)

logging.info("Compare and get service principals to be synced...")

sync_sps = ows.match_service_principals()

print("#" * 80)
print("SERVICE PRINCIPALS TO SYNC")
print("#" * 80)
for sp in sync_sps:
    print(f"{sp.azure_display_name}, {sp.status}")
    if sp.status != SyncStatus.TO_DELETE:
        sp.sync_sp_to_db(force_delete)

logging.info("Compare and get groups to be synced...")

sync_groups = ows.match_groups()
print("#" * 80)
print("GROUPS TO SYNC:")
print("#" * 80)
for group in sync_groups:
    print(f"{group.azure_display_name}, {group.status}")

    if group.status == SyncStatus.TO_ADD:
        group.sync_group_add_empty()

for group in sync_groups:
    if group.status != SyncStatus.TO_DELETE:
        group.sync_group_update(force_delete)

databricks_identity.write_json()



"""
This module is for Databricks Identities objects
"""

import contextlib
import json
from datetime import datetime

from databricks.sdk import AccountClient
from databricks.sdk.service.iam import (
    Group,
    Patch,
    PatchOp,
    PatchSchema,
    ServicePrincipal,
    User,
)


class DatabricksIdentity:
    def __init__(self, account_client: AccountClient, json_path: str):
        """
        Args:
            account_client (databricks.sdk.AccountClient):
                Object containing information needed for authentication in databricks
            relations: A JSON object storing relations between a databricks object
                       and an Azure AD object
            all_databricks_objects: A JSON object storing all databricks objects
            json_path: The json file path (in the root directory)
        """
        self.account_client = account_client
        self.json_path = json_path
        self.relations = self.read_json()
        self.all_databricks_objects = self.get_all_databricks_objects()

    def read_json(self) -> dict:
        """
        This method read syncstates.json file which stores the relations between
        Databricks objects and Azure AD objects. The file is stored in root directory

        Returns:
            dict: a dict of sync states
        """
        with open(self.json_path) as f:
            return json.load(f)

    def write_json(self):
        """
        This method writes the updated relations between
        Databricks objects and Azure AD objects to syncstates.json file
        The file is stored in root directory
        """
        self.relations["lastSyncTime"] = str(datetime.now())
        with open(self.json_path, "w") as outfile:
            outfile.write(json.dumps(self.relations, indent=4))

    # Users
    def fetch_databricks_users(self) -> list[User]:
        """
        This method returns a list of databricks users

        Returns:
            list: a list of all users in databricks
        """
        return list(self.account_client.users.list())

    # def fetch_databricks_user(self, id: str) -> User:
    #     """
    #     Return single user with id

    #     Args:
    #         id (str): databricks id of user
    #     Returns:
    #         databricks.sdk.service.iam.User: user with id
    #     """
    #     return self.account_client.users.get(id)

    def add_databricks_user(
        self, display_name: str, user_name: str, external_id: str
    ) -> User:
        """
        This method adds a user to databricks and returns that user.

        Args:
            display_name (str): Name used to represent the user in databricks
            user_name (str): email of the user
            external_id (str): Azure entra Id of user

        Returns:
            databricks.sdk.service.iam.User:
                databricks user item representing the created user
        """
        new_user = self.account_client.users.create(
            display_name=display_name, user_name=user_name, external_id=external_id
        )
        # external_id won't be updated as it's not supported now
        self.relations["users"][new_user.id] = external_id
        self.all_databricks_objects["users"][new_user.id] = {
            "display_name": new_user.display_name,
            "external_id": external_id,
            "user_name": new_user.user_name,
        }

        return new_user

    def update_databricks_user(
        self,
        id: str,
        display_name: str = None,
        user_name: str = None,
        external_id: str = None,
    ):
        """
        Update a user in databricks.
        Currently only the display_name and user_name are supported for update.
        If the user_name is updated, the databricks user will be deleted and recreated.
        This is done because the databricks API doesn't provide an option to update the
         user_name or email.
        All attributes that don't relate to the user_name are copied over to the new
        user.

        Args:
            id (str): databricks id of the user
            display_name (str): Name used to represent the user in databricks
            user_name (str): email of the user
            external_id (str): azure id synced from Azure
        """
        if user_name:
            old_user = self.account_client.users.get(id)
            old_user.user_name = user_name
            self.account_client.users.delete(id)
            self.account_client.users.create(
                active=old_user.active,
                display_name=old_user.display_name,
                entitlements=old_user.entitlements,
                groups=old_user.groups,
                roles=old_user.roles,
                schemas=old_user.schemas,
                user_name=old_user.user_name,
            )
        else:
            operations = []
            if display_name:
                # Note that the corresponding family_name and given_name
                # will be updated accordingly
                operations.append(Patch(PatchOp.REPLACE, "displayName", display_name))
                self.all_databricks_objects["users"][id]["display_name"] = display_name
            if external_id:
                # the following won't work as it's not supported by Databricks API now
                operations.append(Patch(PatchOp.REPLACE, "externalId", external_id))
                self.all_databricks_objects["users"][id]["external_id"] = external_id

            # We don't update self.relations["users"] here because we want to
            # store relations regardless of whether we perform an update action

            self.account_client.users.patch(
                id,
                operations=operations,
                schemas=[PatchSchema.URN_IETF_PARAMS_SCIM_API_MESSAGES_2_0_PATCH_OP],
            )

    def delete_databricks_user(self, id: str):
        """
        Delete a user in databricks based on it's Id.

        Args:
            id (str): databricks id of the user to be deleted
        """
        self.account_client.users.delete(id=id)
        with contextlib.suppress(KeyError):
            del self.relations["users"][id]
            del self.all_databricks_objects["users"][id]

    # Groups
    def fetch_databricks_groups(self) -> list[Group]:
        """
        This method returns a list of databricks groups

        Returns:
            List: list of groups that exist in databricks
        """
        return list(self.account_client.groups.list())

    # def fetch_databricks_group(self, id: str) -> Group:
    #     """
    #     Return single group with id

    #     Args:
    #         id (str): databricks id of group
    #     Returns:
    #         databricks.sdk.service.iam.Group: group with id
    #     """
    #     return self.account_client.groups.get(id)

    def add_databricks_group(self, display_name: str, external_id: str) -> Group:
        """
        Add a new group without any members in databricks.

        Args:
            display_name (str): Name used to represent the group
            external_id (str): Azure Entra Id of group

        Returns:
            databricks.sdk.service.iam.Group:
                databricks group item representing the created group
        """
        new_group = self.account_client.groups.create(
            display_name=display_name, external_id=external_id
        )

        self.relations["groups"][new_group.id] = external_id
        self.all_databricks_objects["groups"][new_group.id] = {
            "display_name": new_group.display_name,
            "external_id": external_id,
            "members": None,
        }

        return new_group

    # def replace_databricks_group(
    #     self,
    #     id: str,
    #     display_name: str = None,
    #     member_ids: list[str] = None,
    #     external_id: str = None,
    # ):
    #     """
    #     Replace a databricks group.
    #     Currently only the display name and members are supported.

    #     Args:
    #         id (str): databricks id of the group to be updated
    #         display_name (str): Name used to represent the group
    #         member_ids (List[str]):
    #             List of databricks ids of the users that belong to the group
    #     """
    #     if member_ids is not None:
    #         complex_members = self.get_list_of_complex_values(member_ids)
    #     else:
    #         complex_members = self.get_current_members_of_group(id)
    #     self.account_client.groups.update(
    #         id, display_name=display_name, members=complex_members
    #     )

    def update_databricks_group(
        self,
        id: str,
        display_name: str = None,
        member_ids: list[str] = None,
        external_id: str = None,
    ):
        """
        Partially update a databricks group.
        Args:
            id (str): databricks id of the group to be updated
            display_name (str): Name used to represent the group
            member_ids (List[str]):
                List of databricks ids of the users that belong to the group
            external_id: external id of the group
        """
        operations = []
        if member_ids:
            json_members = self.get_list_of_group_members(member_ids)
            operations.append(Patch(PatchOp.REPLACE, "members", json_members))
            self.all_databricks_objects["groups"][id]["members"] = json_members

        if display_name:
            operations.append(Patch(PatchOp.REPLACE, "displayName", display_name))
            self.all_databricks_objects["groups"][id]["display_name"] = display_name

        if external_id:
            operations.append(Patch(PatchOp.REPLACE, "externalId", external_id))
            self.all_databricks_objects["groups"][id]["external_id"] = external_id

        # We don't update self.relations["groups"] here because we want to
        # store relations regardless of whether we perform an update action

        self.account_client.groups.patch(
            id,
            operations=operations,
            schemas=[PatchSchema.URN_IETF_PARAMS_SCIM_API_MESSAGES_2_0_PATCH_OP],
        )

    def delete_databricks_group(self, id: str):
        """
        Delete a group in databricks based on its Id.

        Args:
            id (str): databricks id of the group to be deleted
        """
        self.account_client.groups.delete(id=id)
        with contextlib.suppress(KeyError):
            del self.relations["groups"][id]
            del self.all_databricks_objects["groups"][id]

    # def get_current_members_of_group(self, id: str) -> list[ComplexValue]:
    #     """
    #     Return the members of a group.

    #     Args:
    #         id (str): databricks id of the group that you want to get the members of

    #     Returns:
    #         List (databricks.sdk.service.iam.ComplexValue):
    #             list of complexValues representing the users
    #     """
    #     return self.account_client.groups.get(id=id).members

    # Service Principals
    def fetch_databricks_sps(self) -> list[ServicePrincipal]:
        """
        This method returns a list of databricks service principals
        """
        all_sps = self.account_client.service_principals.list()
        return list(all_sps)

    # def fetch_single_databricks_sps(self, id: str) -> ServicePrincipal:
    #     """
    #     Return single service principal with id

    #     Args:
    #         id (str): databricks id of service principal
    #     Returns:
    #         databricks.sdk.service.iam.ServicePrincipal: service principal with id
    #     """
    #     return self.account_client.service_principals.get(id)

    def add_databricks_sps(
        self, application_id: str, display_name: str, azure_id: str
    ) -> ServicePrincipal:
        """
        Create a new service principal in databricks
        Return: A ServicePrincipal object
        """
        new_sps = self.account_client.service_principals.create(
            active=True,
            display_name=display_name,
            application_id=application_id,
            external_id=azure_id,
        )

        self.relations["servicePrincipals"][new_sps.id] = azure_id
        self.all_databricks_objects["sps"][new_sps.id] = {
            "display_name": new_sps.display_name,
            "application_id": new_sps.application_id,
            "external_id": new_sps.azure_id,
        }

        return new_sps

    def update_databricks_sps(
        self,
        id: str,
        display_name: str = None,
        external_id: str = None,
    ) -> ServicePrincipal:
        """
        Update a service principal in databricks
        """
        operations = []

        if display_name:
            operations.append(Patch(PatchOp.REPLACE, "displayName", display_name))
            self.all_databricks_objects["sps"][id]["display_name"] = display_name

        if external_id:
            operations.append(Patch(PatchOp.REPLACE, "externalId", external_id))
            self.all_databricks_objects["sps"][id]["external_id"] = external_id

        self.account_client.service_principals.patch(
            id,
            operations=operations,
            schemas=[PatchSchema.URN_IETF_PARAMS_SCIM_API_MESSAGES_2_0_PATCH_OP],
        )

    def delete_databricks_sps(self, id: str):
        """
        Delete a Service Principal in databricks based on it's Id.

        Args:
            id (str): databricks id of the user to be deleted
        """
        self.account_client.service_principals.delete(id=id)
        with contextlib.suppress(KeyError):
            del self.relations["servicePrincipals"][id]
            del self.all_databricks_objects["sps"][id]

    # General
    def get_list_of_group_members(self, members: list[str]) -> list[dict]:
        """
        Convert a list of id's to a list of json objects.
        """
        li = []
        if len(members) == 0:
            return li
        for member_id in members:
            li.append(
                {
                    "value": member_id,
                }
            )
        return li

    def get_all_databricks_objects(self) -> dict:
        """
        This method fetches all databricks identities.
        Return: a dict containing all users, groups, service principals
        """
        databricks_json_objects = {
            "users": {},
            "groups": {},
            "sps": {},
        }
        db_groups = self.fetch_databricks_groups()
        for g in db_groups:
            databricks_json_objects["groups"][g.id] = {
                "display_name": g.display_name,
                "external_id": g.external_id,
                "members": g.members,
            }

        db_users = self.fetch_databricks_users()
        for u in db_users:
            databricks_json_objects["users"][u.id] = {
                "display_name": u.display_name,
                "external_id": u.external_id,
                "user_name": u.user_name,
            }

        db_sps = self.fetch_databricks_sps()
        for s in db_sps:
            databricks_json_objects["sps"][s.id] = {
                "display_name": s.display_name,
                "external_id": s.external_id,
                "application_id": s.application_id,
            }

        return databricks_json_objects



"""
This library is used to parse yaml file
"""

import re

import yaml


def parse_yaml(path):
    with open(path) as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)

    # Check if groups list exists in yaml file
    if "groups" not in data:
        raise TypeError("File doesn't contain mandatory list 'groups'.")

    # Check if all items have the structure of an azure object id
    for item in data["groups"]:
        if not is_azure_object_id(item):
            raise ValueError(f"{item} is not a valid Azure Object Id.")
    if "exclude" in data:
        for item in data["exclude"]:
            if not is_azure_object_id(item):
                raise ValueError(f"{item} is not a valid Azure Object Id.")
    else:
        data["exclude"] = []
    return data["groups"], data["exclude"]


def is_azure_object_id(value: str) -> bool:
    rx = "[A-Za-z0-9]{8}-[A-Za-z0-9]{4}-[A-Za-z0-9]{4}-[A-Za-z0-9]{4}-[A-Za-z0-9]{12}"
    x = re.search(rx, value)
    return x is not None
